@@ -28,6 +28,10 @@
 ** Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 ** --------------------------------------------------------------------------
 */
+#define _POSIX_C_SOURCE
+#define _XOPEN_SOURCE
+#define _GNU_SOURCE
+#define _DEFAULT_SOURCE
 
 #include <sys/types.h>
 #include <sys/param.h>
@@ -169,10 +173,10 @@ atopsar(int argc, char *argv[])
 				break;
 
 			   case 'r':		/* reading of file data ? */
-				strncpy(rawname, optarg, RAWNAMESZ-1);
+				safe_strcpy(irawname, optarg, RAWNAMESZ);
 
-				if (strcmp(rawname, "-") == 0)
-					strcpy(rawname, "/dev/stdin");
+				if (strcmp(irawname, "-") == 0)
+					safe_strcpy(irawname, "/dev/stdin", RAWNAMESZ);
 
 				rawreadflag++;
 				break;
@@ -314,7 +318,7 @@ atopsar(int argc, char *argv[])
 		** select own reportraw-function to be called
 		** by the rawread function
 		*/
-		vis.show_samp  = reportraw;
+		handlers[0].handle_sample  = reportraw;
 
 		for (i=0; i < pricnt; i++)
 		{
