@@ -68,7 +68,10 @@ struct tstat {
 		int	wasinactive;	/* boolean: task inactive	*/
 
 		char	utsname[UTSLEN+1];/* UTS name container or pod  */
-		char	cgpath[CGRLEN];	/* cgroup v2 path name          */
+
+		int	cgroupix;	/* index in devchain -1=invalid */
+					/* lazy filling (parsable/json) */
+		int	ifuture[4];	/* reserved for future use	*/
 	} gen;
 
 	/* CPU STATISTICS						*/
@@ -81,10 +84,7 @@ struct tstat {
 		int	policy;		/* scheduling policy            */
 		int	curcpu;		/* current processor            */
 		int	sleepavg;       /* sleep average percentage     */
-		int	cgcpuweight;	/* cgroup cpu.weight            */
-		int	cgcpumax;	/* cgroup cpu.max percentage    */
-		int	cgcpumaxr;	/* restrictive percentage       */
-		int	ifuture[3];	/* reserved for future use	*/
+		int	ifuture[6];	/* reserved for future use	*/
 		char	wchan[16];	/* wait channel string    	*/
 		count_t	rundelay;	/* schedstat rundelay (nanosec)	*/
 		count_t	blkdelay;	/* blkio delay (ticks)		*/
@@ -119,11 +119,7 @@ struct tstat {
 		count_t vlibs;		/* virtmem libexec  (Kb)     	*/
 		count_t vswap;		/* swap space used  (Kb)     	*/
 		count_t	vlock;		/* virtual locked   (Kb) 	*/
-		count_t	cgmemmax;	/* cgroup memory.max (Kb)	*/
-		count_t	cgmemmaxr;	/* restrictive memory.max (Kb)	*/
-		count_t	cgswpmax;	/* cgroup memory.swap.max (Kb)	*/
-		count_t	cgswpmaxr;	/* restrictive swap.max (Kb)	*/
-		count_t	cfuture[3];	/* reserved for future use	*/
+		count_t	cfuture[7];	/* reserved for future use	*/
 	} mem;
 
 	/* NETWORK STATISTICS						*/
@@ -143,7 +139,7 @@ struct tstat {
 
 	struct gpu {
 		char	state;		// A - active, E - Exit, '\0' - no use
-		char	cfuture[3];	//
+		char	bfuture[3];	//
 		short	nrgpus;		// number of GPUs for this process
 		int32_t	gpulist;	// bitlist with GPU numbers
 
@@ -156,6 +152,7 @@ struct tstat {
 		count_t	memnow;		// current    memory consumption in KiB
 		count_t	memcum;		// cumulative memory consumption in KiB
 		count_t	sample;		// number of samples
+		count_t	cfuture[3];	//
 	} gpu;
 };
 
@@ -172,9 +169,9 @@ struct pinfo {
 ** structure to maintains all deviation info related to one sample
 */
 struct devtstat {
-        struct tstat     *taskall;
-        struct tstat    **procall;
-        struct tstat    **procactive;
+        struct tstat     *taskall;	// all processes (also exited) and threads
+        struct tstat    **procall;	// all processes (also exited)
+        struct tstat    **procactive;	// all processes (also exited) being active
 
 	unsigned long	ntaskall;
         unsigned long	ntaskactive;
